from .mwedittypes import EditTypes

__title__ = "mwedittypes"
__summary__ = "mwedittypes is a package that supports edit diffs and action detection for Wikipedia"
__url__ = "https://github.com/geohci/edit-types"

__version__ = "1.1.5"

__license__ = "MIT License"

__all__ = ["EditTypes"]
